"""
Module with the taxonomy class, that serves to create the 
JSON files with the conversion instructions.

Each time the EBA releases a new taxonomy, the taxonomy_loader.py
module must be run to reflect the changes in the taxonomy.
"""

import argparse
import json
import os
import shutil
import subprocess
from pathlib import Path
from tempfile import TemporaryDirectory
from time import time
from zipfile import ZipFile

from lxml import etree

from xbridge.modules import Module

MODULES_FOLDER = Path(__file__).parent / "modules"
INDEX_PATH = MODULES_FOLDER / "index.json"
DIM_DOM_MAPPING_PATH = MODULES_FOLDER / "dim_dom_mapping.json"


def _extract_specific_files_7z(file_path: Path, target_path: Path):
    cmd = [shutil.which('7z'), 'x', f'-o{target_path}', file_path, '*.json', '*dim-def.xml', '-r']
    if not cmd[0]:
        raise FileNotFoundError("7z command not found, please install 7zip to be able to extract "
                                "it and run the script again")
    process = subprocess.run(cmd, capture_output=True)
    if process.returncode != 0:
        raise ValueError(f"Error extracting 7z file {file_path}")


class Taxonomy:
    """
    Class representing an XBRL taxonomy
    """

    def __init__(self):
        self._modules = []

    @property
    def modules(self):
        """Returns the modules within the taxonomy"""
        return self._modules

    @staticmethod
    def __save_module(module, file_path: str | Path = None):
        """Saves a module to a JSON file"""
        with open(file_path, "w", encoding="UTF-8") as fl:
            json.dump(module.to_dict(), fl)

    @staticmethod
    def _get_dim_dom_mapping(root: etree) -> dict:
        ns = {
            'link': 'http://www.xbrl.org/2003/linkbase',
            'xlink': 'http://www.w3.org/1999/xlink'}
        arcroles = root.xpath(
            '//link:definitionArc[@xlink:arcrole="'
            'http://xbrl.org/int/dim/arcrole/dimension-domain"]',
            namespaces=ns)
        map_dom_mapping = {}
        for element in arcroles:
            dim_locator = element.get('{http://www.w3.org/1999/xlink}from')
            dim = root.xpath(f'//link:loc[@xlink:label = "{dim_locator}"]', namespaces=ns)[0]
            dim = dim.get('{http://www.w3.org/1999/xlink}href'). \
                split("#")[1]. \
                split("_")[1]
            dom_locator = element.get('{http://www.w3.org/1999/xlink}to')
            dom = root.xpath(f'//link:loc[@xlink:label = "{dom_locator}"]', namespaces=ns)[0]
            dom = dom.get('{http://www.w3.org/1999/xlink}href').split("#")[1]
            map_dom_mapping[dim] = dom
        return map_dom_mapping

    def load_modules(self, input_path: str | Path = None):
        """loads the modules in the taxonomy"""
        modules = []
        index = {}

        if not MODULES_FOLDER.exists():
            MODULES_FOLDER.mkdir()

        if isinstance(input_path, str):
            input_path = Path(input_path)

        if not input_path.exists():
            raise FileNotFoundError(f"File {input_path} not found")

        if input_path.suffix not in [".zip", ".7z"]:
            raise ValueError("Input file must be a zip or 7z file")

        if input_path.suffix == ".7z":
            print("Converting 7z to zip")
            start = time()
            self._convert_7z_to_zip(input_path)
            input_path = input_path.with_suffix(".zip")
            end = time()
            elapsed = round(end - start, 3)
            print(f"Conversion done in {elapsed} s")

        with ZipFile(input_path, mode="r") as zip_file:
            for file_path in zip_file.namelist():
                file_path_obj = Path(file_path)
                if str(file_path_obj) == \
                        "www.eba.europa.eu\\eu\\fr\\xbrl\\crr\\dict\\dim\\dim-def.xml":
                    bin_read = zip_file.read(file_path)
                    root = etree.fromstring(bin_read.decode('utf-8'))
                    dim_dom_mapping = self._get_dim_dom_mapping(root)
                    with open(DIM_DOM_MAPPING_PATH, "w", encoding="UTF-8") as fl:
                        json.dump(dim_dom_mapping, fl, indent=4)

                if (
                        file_path_obj.suffix == ".json"
                        and file_path_obj.parent.name == "mod"
                ):
                    print(f"Loading module {file_path_obj.stem.upper()}")
                    start = time()
                    module = Module.from_taxonomy(zip_file, file_path)
                    module_file_name = f"{module.code}_{module.date}.json"
                    module_path = str(MODULES_FOLDER / module_file_name)
                    self.__save_module(module, module_path)
                    index_key = f"http://{module.url[:-4]}xsd"
                    index[index_key] = module_file_name
                    modules.append(module)
                    end = time()
                    elapsed = round(end - start, 3)
                    print(f"Module {module.code.upper()} loaded in {elapsed} s")
        if not modules:
            raise TypeError(
                (
                    "No modules found in the taxonomy. "
                    "Please check that the zip file does not contain "
                    "zip files within it"
                )
            )

        with open(INDEX_PATH, "w", encoding="UTF-8") as fl:
            json.dump(index, fl, indent=4)
        self._modules = modules

    def get_module(self, code: str):
        """Returns the module with the given code"""
        for module in self.modules:
            if module.code == code:
                return module
        raise ValueError(f"Module with code {code} not found in the taxonomy")

    def get_variables_from_module(self, code: str) -> list:
        """Returns the variables from the module with the given code"""

        module = self.get_module(code)

        return module.variables

    @classmethod
    def from_taxonomy(cls, input_path: str | Path):
        """Returns a Taxonomy object from a JSON taxonomy file"""
        input_path = input_path if isinstance(input_path, Path) else Path(input_path)
        obj = cls()
        obj.load_modules(input_path)

        ##TODO:
        # Validate that the assumptions for the EBA architecture of the taxonomies is correct:
        # - Currency reported is not an attribute in the CSV
        # - The parameters expect the values we have
        # - There are no potential conflicts if we drop the previxes for
        #       the key values in the scenarios

    @staticmethod
    def _convert_7z_to_zip(input_path):

        """Converts a 7z file to a zip file"""
        start = time()
        with TemporaryDirectory() as temp_folder:
            _extract_specific_files_7z(input_path, Path(temp_folder))
            extracted = time()
            print(f"7z file extracted in {extracted - start} s")
            with ZipFile(input_path.with_suffix(".zip"), mode="w") as zip_file:
                for root, _, files in os.walk(temp_folder):
                    for file in files:
                        # Compute the relative file path to the root of the temporary directory
                        relative_path = os.path.relpath(
                            os.path.join(root, file), temp_folder
                        )
                        # Add the file to the zip file with its relative path
                        zip_file.write(os.path.join(root, file), arcname=relative_path)
        end = time()
        elapsed = round(end - extracted, 3)
        print(f"Zip file created in {elapsed} s")
        elapsed = round(end - start, 3)
        print(f"7z to zip conversion done in {elapsed} s")


def main():
    """Main function to generate the json files from the taxonomy"""

    parser = argparse.ArgumentParser(description="Xbridge taxonomy loader")

    parser.add_argument(
        "input_path",
        type=str,
        help="Please provide the input 7z or zip file with the taxonomy.",
    )

    args = parser.parse_args()
    Taxonomy.from_taxonomy(args.input_path)


if __name__ == "__main__":
    main()



"""
Module holding the Converter class, which converts from XBRL-XML to XBRL-CSV
taking as input the taxonomy object and the instance object
"""

import copy
import csv
import json
from pathlib import Path
from tempfile import TemporaryDirectory
from zipfile import ZipFile

import pandas as pd

from xbridge.modules import Module
from xbridge.xml_instance import Instance

INDEX_FILE = Path(__file__).parent / "modules" / "index.json"
MAPPING_FILE = Path(__file__).parent / "modules" / "dim_dom_mapping.json"

if not INDEX_FILE.exists():
    raise ValueError(
        "Cannot find the index file for the modules. "
        "Please make sure that the index file and the "
        "JSON files with the mappings exist in the modules folder."
    )

with open(INDEX_FILE, "r", encoding="utf-8") as fl:
    index = json.load(fl)


class Converter:
    """
    Converter different types of files into others, using the EBA :obj:`taxonomy <xbridge.taxonomy.Taxonomy>` and XBRL-instance. Each file is extracted and saved in a temporary directory.
    Then, these files are converted into JSON and again saved in a compressed folder such as ZIP or 7z.

    :obj:`Variables <xbridge.taxonomy.Variable>` are generated by combining the open keys column with the attributes found within the taxonomy :obj:`modules <xbridge.taxonomy.Module>`. Then, variable's dimension is extracted
    from it with the purpose to be used with the context coming from the ``XML_instance``, to create the :obj:`scenario <xbridge.xml_instance.Scenario>`.

    Finally, an inner join is done between the variables created and the values from the facts of the ``XML_instance`` :obj:`context <xbridge.xml_instance.Context>`.

    """

    def __init__(self, instance_path: str | Path) -> None:
        self.instance = Instance(instance_path)
        module_ref = self.instance.module_ref

        if module_ref not in index:
            raise ValueError(f"Module {module_ref} not found in the taxonomy index")

        module_path = Path(__file__).parent / "modules" / index[module_ref]
        self.module = Module.from_serialized(module_path)
        self._reported_tables = []

    def convert(self, output_path: str | Path) -> Path:
        """
        Convert the ``XML Instance`` to a CSV file
        """
        if not output_path:
            raise ValueError("Output path not provided")

        if isinstance(output_path, str):
            output_path = Path(output_path)
        if not self.instance:
            raise ValueError("Instance not provided")

        if self.module is None:
            raise ValueError("Module of the instance file not found in the taxonomy")

        temp_dir = TemporaryDirectory()
        temp_dir_path = Path(temp_dir.name)

        meta_inf_dir = temp_dir_path / "META-INF"
        report_dir = temp_dir_path / "reports"

        meta_inf_dir.mkdir()
        report_dir.mkdir()

        with open(meta_inf_dir / "reports.json", "w", encoding="UTF-8") as fl:
            json.dump(
                {
                    "documentInfo": {
                        "documentType": "http://xbrl.org/PWD/2020-12-09/report-package"
                    }
                },
                fl,
            )

        with open(report_dir / "report.json", "w", encoding="UTF-8") as fl:
            json.dump(
                {
                    "documentInfo": {
                        "documentType": "https://xbrl.org/CR/2021-02-03/xbrl-csv",
                        "extends": [self.module.url],
                    }
                },
                fl,
            )

        self._convert_filing_indicator(report_dir)
        with open(MAPPING_FILE, "r", encoding="utf-8") as fl:
            mapping_dict = json.load(fl)
        self._convert_tables(report_dir, mapping_dict)
        self._convert_parameters(report_dir)

        instance_path = self.instance.path

        if not isinstance(instance_path, str):
            instance_path = str(instance_path)

        instance_name = Path(instance_path).name

        if ".xbrl" in instance_name:
            file_name = instance_name.replace(".xbrl", ".zip")
        elif ".xml" in instance_name:
            file_name = instance_name.replace(".xml", ".zip")
        else:
            file_name = instance_name + ".zip"

        zip_file_path = output_path / file_name

        with ZipFile(zip_file_path, "w") as zip_fl:
            for file in meta_inf_dir.iterdir():
                zip_fl.write(file, arcname=f"META-INF/{file.name}")
            for file in report_dir.iterdir():
                zip_fl.write(file, arcname=f"reports/{file.name}")

        temp_dir.cleanup()

        return zip_file_path

    def _variable_generator(self, table):
        """Returns the dataframe with the CSV file for the table

        :param table: The table we use.

        """

        variable_columns = set(table.variable_columns)
        open_keys = set(table.open_keys)
        attributes = set(table.attributes)
        instance_columns = set(self.instance.instance_df.columns)

        # If any open key is not in the instance, then the table cannot have
        # any datapoint
        if not open_keys.issubset(instance_columns):
            return pd.DataFrame(columns=["datapoint", "value"] + list(open_keys))

        # Determine the not relevant dims
        not_relevant_dims = instance_columns - variable_columns - open_keys - attributes
        not_relevant_dims = not_relevant_dims - {"value", "unit", "decimals"}

        instance_df = copy.copy(self.instance.instance_df)
        for col in ["unit", "decimals"]:
            if col not in attributes and col in instance_df.columns:
                del instance_df[col]

        # Drop datapoints that have non-null values in not relevant dimensions
        # And drop the not relevant columns
        instance_df = instance_df[
            instance_df[list(not_relevant_dims)].isnull().all(axis=1)
        ]
        for dim in not_relevant_dims:
            del instance_df[dim]

        # Do the intersection and drop from datapoints the columns and records
        intersect_cols = variable_columns & instance_columns
        datapoint_df = table.variable_df
        for col in variable_columns - instance_columns:
            datapoint_df = datapoint_df[datapoint_df[col].isnull()]
            del datapoint_df[col]

        # Join the dataframes on the datapoint_columns
        table_df = pd.merge(
            datapoint_df, instance_df, on=list(intersect_cols), how="inner"
        )

        if len(table_df) == 0:
            return pd.DataFrame(columns=["datapoint", "value"])

        for col in intersect_cols:
            del table_df[col]
        open_keys_copy = copy.copy(open_keys)
        for open_key_name in open_keys_copy:
            if open_key_name not in table_df.columns:
                open_keys.remove(open_key_name)
        table_df.dropna(subset=list(open_keys), inplace=True)

        return table_df

    def _convert_tables(self, temp_dir_path, mapping_dict):
        for table in self.module.tables:
            ##Workaround:
            # To calculate the table code for abstract tables, we look whether the name
            # ends with a letter, and if so, we remove the last part of the code
            # Possible alternative: add metadata mapping abstract and concrete tables to
            # avoid doing this kind of corrections
            # Defining the output path and check if the table is reported
            normalised_table_code = table.code.replace("-", ".")
            if normalised_table_code[-1].isalpha():
                normalised_table_code = normalised_table_code.rsplit(".", maxsplit=1)[0]
            if normalised_table_code not in self._reported_tables:
                continue

            datapoints = self._variable_generator(table)
            # Cleaning up the dataframe and sorting it
            datapoints = datapoints.rename(columns={"value": "factValue"})
            #Workaround
            #The enumerated key dimensions need to have a prefix like the one
            #Defined by the EBA in the JSON files. We take them from the taxonomy
            #Because EBA is using exactly those for the JSON files.
            for open_key in table.open_keys:
                dim_name = mapping_dict.get(open_key)
                #For open keys, there are no dim_names (they are not mapped)
                if dim_name and not datapoints.empty:
                    datapoints[open_key] = dim_name + ":" + datapoints[open_key].astype(str)
            datapoints = datapoints.sort_values(by=["datapoint"], ascending=True)
            output_path_table = temp_dir_path / table.url
            datapoints.to_csv(output_path_table, index=False)

    def _convert_filing_indicator(self, temp_dir_path):
        # Workaround;
        # Developed for the EBA structure
        output_path_fi = temp_dir_path / "FilingIndicators.csv"
        filing_indicators = self.instance.filing_indicators

        with open(output_path_fi, "w", newline="", encoding="utf-8") as fl:
            csv_writer = csv.writer(fl)
            csv_writer.writerow(["templateID", "reported"])
            for fil_ind in filing_indicators:
                value = "true" if fil_ind.value else "false"
                csv_writer.writerow([fil_ind.table, value])
                if fil_ind.value:
                    self._reported_tables.append(fil_ind.table)

    def _convert_parameters(self, temp_dir_path):
        # Workaround;
        # Developed for the EBA structure
        output_path_parameters = temp_dir_path / "parameters.csv"
        parameters = {
            "entityID": self.instance.entity,
            "refPeriod": self.instance.period,
            "baseCurrency": self.instance.base_currency,
            "decimalsInteger": 0,
            "decimalsMonetary": (
                self.instance.decimals_monetary
                if self.instance.decimals_monetary
                else 0
            ),
            "decimalsPercentage": (
                self.instance.decimals_percentage
                if self.instance.decimals_percentage
                else 4
            ),
        }
        with open(output_path_parameters, "w", newline="", encoding="utf-8") as fl:
            csv_writer = csv.writer(fl)
            csv_writer.writerow(["name", "value"])
            for k, v in parameters.items():
                csv_writer.writerow([k, v])



"""
XBRL Formatter Setup Configuration.
"""

#  __  __    ___              _        _     __ _
#  \ \/ /   | _ )     _ _    (_)    __| |   / _` |   ___
#   >  <    | _ \    | '_|   | |   / _` |   \__, |  / -_)
#  /_/\_\   |___/   _|_|_   _|_|_  \__,_|   |___/   \___|
# _|"""""|_|"""""|_|"""""|_|"""""|_|"""""|_|"""""|_|"""""|
# "`-0-0-'"`-0-0-'"`-0-0-'"`-0-0-'"`-0-0-'"`-0-0-'"`-0-0-'

project = "eba-xbridge"
eba_taxonomy_version = "3.4"
description = f"XBRL-XML to XBRL-CSV converter for EBA Taxonomy (version {eba_taxonomy_version})"
url = ""
version = "1.0.2rc4"
author = "Antonio Olleros, Jesus Simon, Javier Hernandez"
author_email = "antonio.olleros@meaningfuldata.eu, jesus.simon@meaningfuldata.eu, javier.hernandez@meaningfuldata.eu"
copyright = "2024 MeaningfulData"



"""
Module with the classes related to modules, containing the "instructions" for the conversion.
"""

import copy
import json
from pathlib import Path
from zipfile import ZipFile

import pandas as pd


class Module:
    """Class representing an XBRL Module.

    It has attributes like code, url, taxonomy_code, date, tables
    and input_path, whose main function is to operate with the module and return properties like a specific table from
    the JSON file used as input, an object, a dictionary using the attributes as keys, a module object from a part of
    preprocessed JSON file and the variables that are present in it.

    It is used when taxonomies are loaded to collect the information associated to the tables belonging to the module.

    :param code: The code of the XBRL module.

    :param url: The module reference within the taxonomy.

    :param taxonomy_code: The code of the module within to the taxonomy.

    :param date: Dates contained in the module.

    :param tables: The tables that form the module.

    """

    def __init__(self, code=None, url=None, taxonomy_code=None, date=None, tables=None):
        self.code = code
        self.url = url
        self.taxonomy_code = taxonomy_code
        self.date = date
        self._tables = tables if tables is not None else []
        self.taxonomy_module_path = None

    @property
    def tables(self):
        """Returns the :obj:`tables <xbridge.taxonomy.Table>` defined in the JSON file for the :obj:`module <xbridge.taxonomy.Module>`"""
        return self._tables

    def extract_tables(self, zip_file: ZipFile):
        """Extracts the :obj:`tables <xbridge.taxonomy.Table>` in the JSON files for the :obj:`modules <xbridge.taxonomy.Module>` in the taxonomy"""

        self._tables = []
        bin_read = zip_file.read(self.taxonomy_module_path)

        info = json.loads(bin_read.decode("utf-8"))

        for table_code, table in info["tables"].items():
            if table_code[1:] in ("FI", "FootNotes"):
                continue

            table_url = table["url"]
            table_folder_name = table_code[1:].lower()
            table_folder_name = table_folder_name.replace("-", ".")

            path = self.taxonomy_module_path.split("/mod/")[0]

            table_path = (
                path + "/tab/" + table_folder_name + "/" + table_folder_name + ".json"
            )
            table = Table.from_taxonomy(
                zip_file, table_path, code=table_code[1:], url=table_url
            )

            self.tables.append(table)

    def get_table(self, table_code: str):
        """Returns a :obj:`table <xbridge.taxonomy.Table>` object with the given code"""
        for table in self.tables:
            if table.code_name == table_code:
                return table
        raise ValueError(f"Table {table_code} not found in module {self.code}")

    def to_dict(self):
        """Returns a dictionary"""
        return {
            "code": self.code,
            "url": self.url,
            "taxonomy_code": self.taxonomy_code,
            "date": self.date,
            "tables": [tab.to_dict() for tab in self.tables],
        }

    @classmethod
    def from_taxonomy(cls, zip_file: ZipFile, json_file_path: str):
        """Returns a :obj:`module <xbridge.taxonomy.Module>` object from a part of the JSON file"""

        url_split = json_file_path.split("/")
        taxonomy_code = url_split[7]
        date = url_split[8]

        code = Path(json_file_path).stem

        obj = cls(code=code, url=json_file_path, taxonomy_code=taxonomy_code, date=date)

        obj.taxonomy_module_path = json_file_path

        obj.extract_tables(zip_file)

        return obj

    @classmethod
    def from_serialized(cls, input_path: str | Path):
        """Returns a :obj:`module <xbridge.taxonomy.Module>` object from a JSON file"""
        input_path = input_path if isinstance(input_path, Path) else Path(input_path)
        with open(input_path, "r", encoding="UTF-8") as fl:
            module_dict = json.load(fl)

        tables = module_dict.pop("tables")
        tables = [Table.from_dict(table) for table in tables]

        obj = cls(**module_dict, tables=tables)

        return obj

    @property
    def variables_location(self):
        """Returns a dictionary with the :obj:`variables <xbridge.taxonomy.Variable>`
        and the :obj:`tables <xbridge.taxonomy.Table>` where they are present"""
        variables = {}
        for table in self.tables:
            for variable in table.variables:
                if variable.code not in variables:
                    variables[variable.code] = [table.code]
                else:
                    variables[variable.code].append(table.code)
        return variables

    @property
    def repeated_variables(self):
        """Returns a dictionary with the :obj:`variables <xbridge.taxonomy.Variable>` and the :obj:`tables <xbridge.taxonomy.Table>`
        where they are present, if they are repeated"""
        result = {}
        for k, v in self.variables_location.items():
            if len(v) > 1:
                result[k] = v
        return result

    def __repr__(self) -> str:
        return f"<Module - {self.code}>"


class Table:
    """Class representing an XBRL :obj:`table <xbridge.taxonomy.Table>` as defined in the JSON file.

    Its properties allow to return open keys, variables and attributes from the :obj:`table <xbridge.taxonomy.Table>`. It can also generate a
    variable dataframe or work with one already created.
    Finally, it can return a dictionary using its attributes or a :obj:`Table <xbridge.taxonomy.Table>` object from the preprocessed JSON file.

    It is used when module is loaded to collect the information associated to the variables and open keys
    belonging to the table.

    :param code: The code of the table.

    :param url: The table reference within the module.

    :param open_keys: Open key contained in the table.

    :param variables: the variables that belongs to the table.

    :param attributes: attributes related to the variables that can be extracted from the table.

    :param input_zip_path: Path to the file used as table.
    """

    def __init__(
        self,
        code=None,
        url=None,
        open_keys=None,
        variables=None,
        attributes=None,
        input_zip_path=None,
    ):

        self.table_zip_path = input_zip_path
        self.code = code
        self.url = url
        self._open_keys = open_keys if open_keys is not None else []
        self._variables = variables if variables is not None else []
        self._attributes = attributes if attributes is not None else []
        self._datapoint_df = None

    @property
    def open_keys(self):
        """Returns the open keys for the :obj:`table <xbridge.taxonomy.Table>`"""
        return self._open_keys

    @property
    def variables(self):
        """Returns the :obj:`variable <xbridge.taxonomy.Variable>` for the :obj:`table <xbridge.taxonomy.Table>`"""
        return self._variables

    @property
    def attributes(self):
        """Returns the attributes for the :obj:`table <xbridge.taxonomy.Table>`"""
        return self._attributes

    @property
    def variable_columns(self):
        """
        Returns the columns for the :obj:`variable <xbridge.taxonomy.Variable>` dataframe
        """
        cols = set(self.variable_df.columns)
        cols.remove("datapoint")
        return cols

    @property
    def variable_df(self):
        """
        Returns a dataframe with the :obj:`variable <xbridge.taxonomy.Variable>` and extensional context

        """
        return self._datapoint_df

    def generate_variable_df(self):
        """Returns a dataframe with the :obj:`variable <xbridge.taxonomy.Variable>` and extensional context"""
        variables = []
        for variable in self.variables:
            variable_info = {}
            for dim_k, dim_v in variable.dimensions.items():
                if dim_k not in ("unit", "decimals"):
                    variable_info[dim_k] = dim_v.split(":")[1]
            if "concept" in variable.dimensions:
                variable_info["metric"] = variable.dimensions["concept"].split(":")[1]
                del variable_info["concept"]

            variable_info["datapoint"] = variable.code
            variables.append(copy.copy(variable_info))
        self._datapoint_df = pd.DataFrame(variables)

    def extract_open_keys(self, zip_file: ZipFile):
        """Extracts the open keys for the :obj:`table <xbridge.taxonomy.Table>`"""
        self._open_keys = []
        self._attributes = []

        bin_read = zip_file.read(self.table_zip_path)

        table_json = json.loads(bin_read.decode("utf-8"))
        table_template = table_json["tableTemplates"][self.code]
        for column_name in table_template.get("columns", []):
            if column_name == "unit":
                self._attributes.append(column_name)
            elif column_name not in ("datapoint", "factValue"):
                self._open_keys.append(column_name)

    def extract_variables(self, zip_file: ZipFile):
        """Extract the :obj:`variable <xbridge.taxonomy.Variable>` for the :obj:`table <xbridge.taxonomy.Table>`"""
        self._variables = []
        bin_read = zip_file.read(self.table_zip_path)

        table_json = json.loads(bin_read.decode("utf-8"))
        if self.code in table_json["tableTemplates"]:
            variables_dict = table_json["tableTemplates"][self.code]["columns"][
                "datapoint"
            ]["propertyGroups"]
            for elto_k, elto_v in variables_dict.items():
                datapoint = Variable.from_taxonomy(elto_k, elto_v)
                self._variables.append(datapoint)

    def to_dict(self):
        """Returns a dictionary for the :obj:`table <xbridge.taxonomy.Table>`"""
        return {
            "code": self.code,
            "url": self.url,
            "open_keys": self.open_keys,
            "variables": [var.to_dict() for var in self.variables],
            "attributes": self.attributes,
        }

    @classmethod
    def from_taxonomy(cls, zip_file: ZipFile, table_path: str, code: str, url: str):
        """Returns a :obj:`table <xbridge.taxonomy.Table>` object from a part of the preprocessed JSON file"""
        obj = cls(code=code, url=url)
        obj.table_zip_path = table_path

        obj.extract_open_keys(zip_file)
        obj.extract_variables(zip_file)

        # obj.generate_datapoint_df()

        return obj

    @classmethod
    def from_dict(cls, table_dict):
        """Returns a :obj:`table <xbridge.taxonomy.Table>` object from a dictionary"""
        variables = table_dict.pop("variables")
        variables = [Variable.from_dict(variable) for variable in variables]

        obj = cls(**table_dict, variables=variables)
        obj.generate_variable_df()

        return obj

    def __repr__(self) -> str:
        return f"<Table - {self.code}>"


class Variable:
    """Class representing a variable as represented in the JSON files. Can return or extract the `dimension <https://www.xbrl.org/guidance/xbrl-glossary/#:~:text=a%20taxonomy.-,Dimension,-A%20qualifying%20characteristic>`_
    of the :obj:`variable <xbridge.taxonomy.Variable>`, create a dictionary using its attributes as keys or return a variable object from the
    preprocessed JSON file.


    :param code: The code of the variable.

    :param dimensions: the `dimensions <https://www.xbrl.org/guidance/xbrl-glossary/#:~:text=a%20taxonomy.-,Dimension,-A%20qualifying%20characteristic>`_ of the variable.

    :param attributes: The attributes related to the variable.

    """

    def __init__(self, code=None, dimensions=None, attributes=None):
        self.code = code
        self._dimensions = dimensions
        self._attributes = attributes

    @property
    def dimensions(self):
        """Returns the `dimensions <https://www.xbrl.org/guidance/xbrl-glossary/#:~:text=a%20taxonomy.-,Dimension,-A%20qualifying%20characteristic>`_ of a variable"""
        return self._dimensions

    def extract_dimensions(self, datapoint_dict):
        """Extracts the `dimensions <https://www.xbrl.org/guidance/xbrl-glossary/#:~:text=a%20taxonomy.-,Dimension,-A%20qualifying%20characteristic>`_ for the variable"""
        self._dimensions = datapoint_dict["dimensions"]
        if "decimals" in datapoint_dict:
            self._attributes = datapoint_dict["decimals"]

    def to_dict(self):
        """Returns a dictionary with the attributes"""
        return {
            "code": self.code,
            "dimensions": self.dimensions,
            "attributes": self._attributes,
        }

    @classmethod
    def from_taxonomy(cls, variable_id, variable_dict):
        """Returns a :obj:`variable <xbridge.taxonomy.Variable>` object from a part of the preprocessed JSON file"""
        obj = cls(code=variable_id)
        obj.extract_dimensions(variable_dict)

        return obj

    @classmethod
    def from_dict(cls, variable_dict):
        """Returns a :obj:`variable <xbridge.taxonomy.Variable>` object from a dictionary"""
        modified_dimensions = {}
        for k, v in variable_dict["dimensions"].items():
            if ":" in k:
                k = k.split(":")[1]
                modified_dimensions[k] = v
            else:
                modified_dimensions[k] = v
        modified_dict = variable_dict.copy()
        modified_dict["dimensions"] = modified_dimensions
        obj = cls(**modified_dict)
        # print(modified_dict)
        return obj

    def __repr__(self) -> str:
        return f"<Variable - {self.code}>"
